#! /usr/bin/env python

#############################################################################
#
# (C) 2021 Cadence Design Systems, Inc. All rights reserved worldwide.
#
# This sample script is not supported by Cadence Design Systems, Inc.
# It is provided freely for demonstration purposes only.
# SEE THE WARRANTY DISCLAIMER AT THE BOTTOM OF THIS FILE.
#
#############################################################################

"""
This module provides a wrapper around TCP communication of a client with
a Glyph Server.

Example:
    from pointwise import GlyphClient, GlyphError

    glf = GlyphClient()
    if glf.connect():
        try:
            result = glf.eval('pw::Application getVersion')
            print('Pointwise version is {0}'.format(result))
        except GlyphError as e:
            print('Error in command {0}\n{1}'.format(e.command, e))

    elif glf.is_busy():
        print('The Glyph Server is busy')
    elif glf.auth_failed():
        print('Glyph Server authentication failed')
    else:
        print('Failed to connect to the Glyph Server')
"""

import os, time, socket, struct, errno, sys, re, platform

class GlyphError(Exception):
    """ This exception is raised when a command passed to the Glyph Server
        encounters an error. 
    """
    def __init__(self, command, message, *args, **kwargs):
        Exception.__init__(self, message, *args, **kwargs)
        self.command = command


class NoLicenseError(Exception):
    """ This exception is raised when a Glyph server subprocess could
        not acquire a valid Pointwise license.
    """
    def __init__(self, message, *args, **kwargs):
        Exception.__init__(self, message, *args, **kwargs)


class GlyphClient(object):
    """ This class is a wrapper around TCP client communications with a
        Glyph Server.  Optionally, it can start a batch Glyph Server process
        as a subprocess (which consumes a Pointwise license). To start
        a server subprocess, initialize the GlyphClient with port = 0.
        For Windows platforms, the default program to run is 'tclsh'. For
        Linux and Mac OS platforms, the default program is 'pointwise -b'.
        (Optionally, the 'prog' argument may be specified to indicate the
        program or shell script to run as the batch Glyph server process.
        This is typically used in environments where multiple versions of
        Pointwise are installed.)
    """

    def __init__(self, port=None, auth=None, version=None, host='localhost', \
            callback=None, prog=None, timeout=10):
        """ Initialize a GlyphClient object

            Args:
                port (str): the port number of the Glyph Server to connect to.
                    If the port is not given, it defaults to the environment
                    variable PWI_GLYPH_SERVER_PORT, or 2807 if not defined.
                    If port is 0, a Pointwise batch subprocess will be started
                    if a license is available.
                auth (str): the authorization code of the Glyph Server.  If the
                    auth is not given, it defaults to the environment variable
                    PWI_GLYPH_SERVER_AUTH, or an empty string if not defined.
                version (str): the Glyph version number to use for
                    compatibility. The string should be of the form X.Y.Z
                    where X, Y, and Z are positive integer values. The Y and
                    Z values are optional and default to a zero value. A blank
                    value always uses the current version. 
                host (str): the host name of the Glyph Server to connect to.
                    The default value is 'localhost'.
                callback (func): a callback function that takes a string
                    parameter that represents the response messages from the
                    Glyph Server. The default callback is None.
                prog (str): The batch executable program to use when port is 0.
                    If prog is None, 'tclsh' will be used on Windows platforms
                    and 'pointwise -b' will be used for Linux/macOS platforms.
                    Otherwise the given program will be used to execute the
                    Pointwise batch mode.
                timeout (float): The number of seconds to continue to try to
                    connect to the server before giving up. The default timeout
                    is 10 seconds.
        """
        self._port = port
        self._auth = auth
        self._host = host
        self._version = version
        self._timeout = timeout

        self._socket = None
        self._busy = False
        self._auth_failed = False
        self._serverVersion = None

        if self._port is None:
            self._port = os.environ.get('PWI_GLYPH_SERVER_PORT', '2807')

        if self._auth is None:
            self._auth = os.environ.get('PWI_GLYPH_SERVER_AUTH', '')

        if self._port == 0:
            self._startServer(callback, prog)


    def __del__(self):
        if hasattr(self, "_server") and self._server is not None:
            self._server.stdout = None
            self._server.stderr = None
            self._server.stdin = None
        self.close()


    def __eq__(self, other):
        return self is other


    def __enter__(self):
        """ When using GlyphClient as a context manager, connect and return self
        """
        if not self.connect():
            if self._auth_failed:
                raise GlyphError('connect', 'Glyph Server AUTH failed')
            elif self._busy:
                raise GlyphError('connect', 'Glyph Server BUSY')
            else:
                raise GlyphError('connect', 'Could not connect to Glyph Server')
        return self


    def __exit__(self, *args):
        """ When using GlyphClient as a context manager, disconnect """
        self.close()

    def __str__(self):
        s = 'GlyphClient(' + str(self._host) + '@' + str(self._port) + \
            ') connected=' + str(self.is_connected())
        if self.is_connected():
            s = s + ' Server=' + self._serverVersion
        return s


    def connect(self, retries=None):
        """ Connect to a Glyph server at the given host and port.

            Args:
                retries (int): the number of times to retry the connection
                    before giving up. DEPRECATED: if not None, retries will be
                    used to determine a suitable value for timeout.

            Returns:
                bool: True if successfully connected, False otherwise.  If an
                    initial connection is made, but the Glyph Server is busy,
                    calling is_busy() will return True.
        """
        self._closeSocket()

        self._busy = False
        self._auth_failed = False

        timeout = self._timeout
        if retries is not None:
            timeout = 0.1 * retries

        start = time.time()
        while self._socket is None and time.time() - start < timeout:
            try:
                self._socket = self._connect(self._host, self._port)
            except:
                self._socket = None
            if self._socket is None:
                time.sleep(0.1) # sleep for a bit before retry

        if self._socket is None:
            return False

        self._send('AUTH', self._auth)

        response = self._recv()
        if response is None:
            self._socket.close()
            self._socket = None
            return False

        rtype, payload = response

        self._auth_failed = (rtype == 'AUTHFAIL')
        self._busy = (rtype == 'BUSY')

        if rtype != 'READY':
            self.close()
        else:
            problem = None
            if self._version is not None:
                try:
                    self.control('version', self._version)
                except Exception as excCode:
                    problem = excCode
            self._serverVersion = self.eval('pw::Application getVersion')
            if problem is not None:
                m = re.search('Pointwise V(\d+).(\d+)', self._serverVersion)
                if len(m.groups()) == 2:
                    try:
                        serverVers = int(m.groups()[0]) * 10 + \
                            int(m.groups()[1])
                    except:
                        serverVers = 0
                    if serverVers >= 183:
                        self.close()
                        raise problem

        return self._socket is not None


    def is_busy(self):
        """ Check if the reason for the last failed attempt to connect() was
            because the Glyph Server is busy.

            Returns:
                bool: True if the last call to connect failed because the
                    Glyph Server was busy, False otherwise.
        """
        return self._busy 


    def auth_failed(self):
        """ Check if the reason for the last failed attempt to connect() was
            because the Glyph Server authentication failed.

            Returns:
                bool: True if the last call to connect failed because the
                    Glyph Server authentication failed, False otherwise.
        """
        return self._auth_failed


    def eval(self, command):
        """ Evaluate a Glyph command on the Glyph Server, including nested
            commands and variable substitution.

            Args:
                command (str): A Glyph command to be evaluated on the
                Glyph Server.

            Returns:
                str: The result from evaluating the command on the Glyph Server.

            Raises:
                GlyphError: If the command evaluation on the Glyph Server
                resulted in an error.
        """
        return self._sendcmd('EVAL', command)


    def command(self, command):
        """ Execute a Glyph command on the Glyph Server, with no support
            for nested commands and variable substitution.

            Args:
                command (str): A JSON encoded string of an array of a
                    Glyph command and the parameters to be executed.

            Returns:
                str: The JSON encoded string result from executing the
                    command on the Glyph Server.

            Raises:
                GlyphError: If the command execution on the Glyph Server
                    resulted in an error.
        """
        return self._sendcmd('COMMAND', command)


    def control(self, setting, value=None):
        """ Send a control message to the Glyph Server

            Args:
                setting (str): The name of the control setting
                value (str): The value to set the setting to. If none the value
                    will only be queried. 

            Returns:
                str: The current value of the control setting

            Raises:
                GlyphError: If the control setting or the value is invalid.
        """
        command = setting
        if value is not None:
            command += '=' + value
        return self._sendcmd('CONTROL', command)
    
    
    def get_glyphapi(self):
        """ Creates and returns a GlyphAPI object for this client
            
            Returns: 
                GlyphAPI object for automatic Glyph command execution
                through this client connection
        """
        from pointwise import glyphapi
        if not self.is_connected():
            self.connect()

        if self.is_connected():
            return glyphapi.GlyphAPI(glyph_client=self)
        else:
            raise GlyphError('GlyphAPI',
                    'The client is not connected to a Glyph Server')
    
    
    def is_connected(self):
        """ Check if there is a connection to the Glyph Client
            
            Returns:
                True if there is a valid connection, False otherwise.
        """
        result = False
        try:
            result = self._socket is not None and self.ping()
        except:
            result = False
        return result    
    
    
    def ping(self):
        """ Ping the Glyph Server

            Returns:
                bool: True if the the Glyph Server was successfully pinged,
                    False otherwise.
        """
        result = False
        try:
            result = self._sendcmd('PING', '') == "OK"
        except:
            result = False
        return result


    def close(self):
        """ Close the connection with the Glyph server """

        if hasattr(self, "_server") and self._server is not None and \
                hasattr(self, "_socket") and self._socket is not None:
            try:
                # request graceful shutdown of server
                self.eval("exit")
            except socket.error as err:
                pass

        self._closeSocket()

        if hasattr(self, "_server") and self._server is not None:
            # safe to call even if server has already shut down
            self._server.terminate()
            self._server.wait()

            # resource warnings may occur if pipes are not closed from the
            # client side (typically on Windows)
            if self._server.stdout is not None:
                try:
                    self._server.stdout.close()
                    self._server.stdout = None
                except IOError:
                    pass

            # wait for the pipe reader thread to finish
            if self._othread is not None:
                self._othread.join(0.5)
                del self._othread
                self._othread = None

            if self._server.stdin is not None:
                try:
                    self._server.stdin.close()
                    self._server.stdin = None
                except IOError:
                    pass

            if self._server.stderr is not None:
                try:
                    self._server.stderr.close()
                    self._server.stderr = None
                except IOError:
                    pass

            self._server = None


    def disconnect(self):
        """ Close the connection with the Glyph server """
        self.close()


    def _connect(self, host, port):
        """ Helper function for connecting a socket to the given host and port
        """
        # try to connect using both IPv4 and IPv6
        s = None
        for res in socket.getaddrinfo(host, port, socket.AF_UNSPEC,
                socket.SOCK_STREAM):
            af, socktype, proto, canonname, sa = res
            try:
                s = socket.socket(af, socktype, proto)
                s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            except (socket.error, OSError) as msg:
                s = None
                continue
            try:
                s.connect(sa)
            except (socket.error, OSError) as msg:
                s.close()
                s = None
                continue
            break
        return s


    def _sendcmd(self, type, command):
        """ Helper function for sending a command to the Glyph Server """
        if self._socket is None:
            raise GlyphError(command,
                    'The client is not connected to a Glyph Server')

        try:
            self._send(type, command)
            response = self._recv()
        except socket.error as e:
            # Python 3: BrokenPipeError
            if e.errno == errno.EPIPE:
                raise GlyphError(command,
                        'The Glyph Server connection is closed')
            else:
                raise

        if response is None:
            if not self._socket.fileno() == -1:
                # socket is closed, assume command ended server session
                return None
            raise GlyphError(command, 'No response from the Glyph Server')

        type, payload = response
        if type != 'OK':
            raise GlyphError(command, payload)

        return payload


    def _send(self, type, payload):
        """ Helper function for sending a message on the socket """
        message_bytes = type.encode('utf-8').ljust(8) + payload.encode('utf-8')
        message_length = struct.pack('!I', len(message_bytes))
        self._socket.sendall(message_length)
        self._socket.sendall(message_bytes)


    def _recv(self):
        """ Helper function for receiving a message on the socket """
        message_length = self._recvall(4)
        if message_length is None:
            return None

        # Python 2: 'bytes' is an alias for 'str'
        # Python 3: 'bytes' is an immutable bytearray
        message_length = struct.unpack('!I', bytes(message_length))[0]
        if message_length == 0:
            return ('', '')

        message_bytes = self._recvall(message_length)
        if message_bytes is None:
            return None

        # Python 2: convert decoded bytes (type 'unicode') to string (type
        # 'str')
        type = str(message_bytes[0:8].decode('utf-8')).strip()
        payload = str(message_bytes[8:].decode('utf-8'))
        return (type, payload)


    def _recvall(self, size):
        """ Helper function to recv size bytes or return None if EOF is hit """
        data = bytearray()
        while len(data) < size:
            packet = self._socket.recv(size - len(data))
            if not packet:
                return None
            data.extend(packet)
        return data

    def puts(self, *args):
        self.eval("puts {%s}" % ' '.join(args))


    def _closeSocket(self):
        """ Close the connection with the Glyph server """
        if self._socket:
            if not self._socket.fileno() == -1:
                try:
                    self._socket.shutdown(socket.SHUT_RDWR)
                except:
                    None
                self._socket.close()
                del self._socket
            self._socket = None


    def _startServer(self, callback, prog):
        """ Create a server if possible on any open port """
        self._server = None
        self._othread = None
        try:
            if prog is None:
                if platform.system() == 'Windows':
                    prog = ['tclsh']
                else:
                    prog = ['pointwise', '-b']
            elif isinstance(prog, tuple):
                prog = list(prog)
            elif not isinstance(prog, list):
                prog = [prog]

            import shutil
            if not hasattr(shutil, 'which'):
                # Python 3: shutil.which exists
                shutil.which = __which__

            target = shutil.which(prog[0])

            if target is None:
                raise GlyphError('server', '%s not found in path' % prog[0])

            prog[0] = target

            # find an open port
            try:
                tsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                tsock.settimeout(0)
                tsock.bind(('', 0))
                self._port = tsock.getsockname()[1]
            finally:
                tsock.close()
                time.sleep(0.1)

            if callback is None:
                def __default_callback__(*args):
                    pass
                callback = __default_callback__

            # start the server subprocess
            import subprocess
            self._server = subprocess.Popen(prog, stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

            self._server.stdin.write(bytearray((
                "if [catch {package require PWI_Glyph} server_ver] {\n" +
                "  puts {Could not load Glyph package. Ensure that the " +
                "Pointwise version of tclsh is the only one in your PATH.}\n" +
                "  puts $server_ver\n" +
                "  exit\n" +
                "}\n" +
                "if { [package vcompare $server_ver 2.18.2] < 0 } {" +
                "  puts {Server must be 18.2 or higher}\n"
                "  exit\n"
                "}\n" +
                "puts \"GLYPH_SERVER_VERSION: $server_ver\"\n" +
                "pw::Script setServerPort %d\n" +
                "puts \"Server: [pw::Application getVersion]\"\n" +
                "pw::Script processServerMessages -timeout %s\n" +
                "puts \"Server: Subprocess completed.\"\n") %
                (self._port, str(int(self._timeout))), "utf-8"))
            self._server.stdin.flush()

            ver_re = re.compile(r"GLYPH_SERVER_VERSION: (\d+\.\d+\.\d+)")
            lic_re = re.compile(r".*unable to obtain a license.*")

            # process output from the server until we can determine
            # that the server is running and a valid license was obtained
            ver = str(self._server.stdout.readline().decode('utf-8'))
            ver_match = ver_re.match(ver)
            lic_match = lic_re.match(ver)

            # Read the server output, looking for a special line with the
            # server version, or a line with a known license failure message,
            # or until EOF. Print all output preceding any of these conditions.
            if ver_match is None and lic_match is None:
                # Note: not strictly Python 2 compatible, but deemed acceptable
                print(ver)
                for ver in iter(self._server.stdout.readline, b''):
                    ver = str(ver.decode('utf-8'))
                    ver_match = ver_re.match(ver)
                    if ver_match is not None:
                        break
                    lic_match = lic_re.match(ver)
                    if lic_match is not None:
                        break
                    print(ver)

            if lic_match is not None or ver_match is None:
                # license or other failure
                for line in iter(self._server.stdout.readline, b''):
                    callback(str(line.decode('utf-8')))
                self.close()
                if lic_match is not None:
                    raise NoLicenseError(ver)
                else:
                    raise GlyphError('server', ver)

            # capture stdout/stderr from the server
            import threading
            class ReaderThread(threading.Thread):
                def __init__(self, ios, callback):
                    threading.Thread.__init__(self)
                    self._ios = ios
                    self._error = None
                    self._cb = callback
                    self.daemon = True

                def run(self):
                    try:
                        for line in iter(self._ios.readline, b''):
                            # Python 2: convert decoded bytes to 'str'
                            self._cb(str(line.decode('utf-8')))
                    except Exception as ex:
                        self._error = str(ex)
                    self._ios = None

            self._othread = ReaderThread(self._server.stdout, callback)
            self._othread.start()
        except Exception as ex:
            if self._server is not None:
                self._server.kill()
                self._server.wait()
                self._server = None
            if self._othread is not None:
                self._othread.join(0.5)
                if self._othread._error is not None:
                    callback(self._othread._error)
            raise


def __which__(cmd, mode=os.F_OK | os.X_OK, path=None):
    """Given a command, mode, and a PATH string, return the path which
       conforms to the given mode on the PATH, or None if there is no such
       file.
       `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
       of os.environ.get("PATH"), or can be overridden with a custom search
       path.

       Courtesy Python 3.3 source code, for Python 2.x backward compatibility.
    """

    # Check that a given file can be accessed with the correct mode.
    # Additionally check that `file` is not a directory, as on Windows
    # directories pass the os.access check.
    def __access_check__(fn, mode):
        return (os.path.exists(fn) and os.access(fn, mode) and
            not os.path.isdir(fn))

    # Short circuit. If we're given a full path which matches the mode
    # and it exists, we're done here.
    if __access_check__(cmd, mode):
        return cmd

    path = (path or os.environ.get("PATH", os.defpath)).split(os.pathsep)

    if sys.platform == "win32":
        # The current directory takes precedence on Windows.
        if os.curdir not in path:
            path.insert(0, os.curdir)

        # PATHEXT is necessary to check on Windows.
        pathext = os.environ.get("PATHEXT", "").split(os.pathsep)
        # See if the given file matches any of the expected path extensions.
        # This will allow us to short circuit when given "python.exe".
        matches = [cmd for ext in pathext if cmd.lower().endswith(ext.lower())]
        # If it does match, only test that one, otherwise we have to try
        # others.
        files = [cmd] if matches else [cmd + ext.lower() for ext in pathext]
    else:
        # On other platforms you don't have things like PATHEXT to tell you
        # what file suffixes are executable, so just pass on cmd as-is.
        files = [cmd]

    seen = set()
    for dir in path:
        dir = os.path.normcase(dir)
        if dir not in seen:
            seen.add(dir)
            for thefile in files:
                name = os.path.join(dir, thefile)
                if __access_check__(name, mode):
                    return name
    return None

#############################################################################
#
# This file is licensed under the Cadence Public License Version 1.0 (the
# "License"), a copy of which is found in the included file named "LICENSE",
# and is distributed "AS IS." TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE
# LAW, CADENCE DISCLAIMS ALL WARRANTIES AND IN NO EVENT SHALL BE LIABLE TO
# ANY PARTY FOR ANY DAMAGES ARISING OUT OF OR RELATING TO USE OF THIS FILE.
# Please see the License for the full text of applicable terms.
#
#############################################################################



#! /usr/bin/env python

#############################################################################
#
# (C) 2021 Cadence Design Systems, Inc. All rights reserved worldwide.
#
# This sample script is not supported by Cadence Design Systems, Inc.
# It is provided freely for demonstration purposes only.
# SEE THE WARRANTY DISCLAIMER AT THE BOTTOM OF THIS FILE.
#
#############################################################################

import numpy as np
import math


"""
Utility classes that mimic the behavior of the Glyph pwu::* utilites.
"""

class GlyphUtility:
    tolerance = 1e-7

class Vector2(object):
    """ Utility functions for two dimensional vectors, which are represented
        as a list of two real values.
    """

    def __init__(self, *args):
        """ Construct a 2D vector, all zeroes by default.  """
        self.vector_ = None
        if len(args) == 0:
            self.vector_ = np.zeros([2])
        elif len(args) == 1:
            if isinstance(args[0], (list, tuple)) and len(args[0]) == 2:
                self.vector_ = np.array([float(i) for i in args[0]])
            elif isinstance(args[0], np.ndarray):
                self.vector_ = args[0]
            elif isinstance(args[0], Vector2):
                self.vector_ = args[0].vector_
        elif len(args) == 2:
            self.vector_ = np.array([float(i) for i in args])

        if self.vector_ is None:
            raise ValueError("Invalid parameter %s" % str(args))
        elif self.vector_.size != 2 or self.vector_.ndim != 1:
            raise ValueError("Vector2 must be a list/tuple of 2 values")

    def __eq__(self, other):
        """ Check for equality of two vectors """
        return self.equal(other)

    def __add__(self, other):
        """ Add two vectors together and return the result as a new Vector2 """
        return Vector2(np.add(tuple(self), tuple(other)))
    
    def __sub__(self, other):
        """ Subtract one vector from another and return the result as a
            new Vector2.
        """
        return Vector2(np.subtract(tuple(self), tuple(other)))
    
    def __mul__(self, other):
        """ Multiply the components of a vector by either the components of
            another vector or a scalar and return the result as a new
            Vector2.
        """
        if isinstance(other, (float, int)):
            other = (other, other)

        return Vector2(np.multiply(tuple(self), tuple(other)))
        
    def __truediv__(self, other):
        """ Divide the components of a vector by either the components of
            another vector or a scalar and return the result as a new
            Vector2.
        """
        if isinstance(other, (float,int)):
            other = (other, other)

        return Vector2(np.true_divide(tuple(self), tuple(other)))

    def __div__(self, other):
        """ Divide the components of a vector by either the components of
            another vector or a scalar and return the result as a new
            Vector2.
        """
        return self.__truediv__(other)

    def __str__(self):
        """ Return a string representation of a Vector2 """
        return str(tuple(self))

    def __iter__(self):
        return (e for e in self.vector_)
 
    def __len__(self):
        return len(self.vector_)

    def index(self, i):
        """ Return a component of a Vector2 """
        return self.vector_[i]
    
    @property
    def x(self):
        """ The X component of a Vector2 """
        return self.vector_[0]

    @x.setter
    def x(self, v):
        """ Set the X component of a Vector2 """
        self.vector_[0] = v
    
    @property
    def y(self):
        """ The X component of a Vector2 """
        return self.vector_[1]

    @y.setter
    def y(self, v):
        """ Set the Y component of a Vector2 """
        self.vector_[1] = v
    
    def equal(self, other, tolerance=GlyphUtility.tolerance):
        """ Check for equality between two vectors within an optional tolerance
        """
        return np.linalg.norm(np.subtract(tuple(self), tuple(other))) \
                <= tolerance
        
    def notEqual(self, other, tolerance=GlyphUtility.tolerance):
        """ Check for inequality between two vectors within an optional
            tolerance
        """
        return not self.equal(other, tolerance)
        
    def add(self, vec2):
        """ Add the components of two vectors and return the result as a new
            Vector2
        """
        return self + vec2
    
    def subtract(self, vec2):
        """ Subtract one vector from another and return the result as a new
            Vector2
        """
        return self - vec2
    
    def negate(self):
        """ Negate the component values and return the result as a new Vector2
        """
        return self * -1
    
    def scale(self, scalar):
        """ Scale the component values by a scalar and return the result as
            a new Vector2
        """
        return self * scalar
    
    def divide(self, scalar):
        """ Divide the component values by a scalar and return the result as
            a new Vector2
        """
        return self / scalar
    
    def multiply(self, vec2):
        """ Multiply the component values of a vector by the component values
            of another vector and return the result as a new Vector2
        """
        return self * vec2
    
    def dot(self, vec2):
        """ Return the scalar dot product of two vectors """
        return np.dot(self.vector_, vec2.vector_)
    
    def normalize(self):
        """ Return the normalized form of a vector as a new Vector2 """
        norm = np.linalg.norm(self.vector_)
        if norm == 0:
            raise ValueError('Vector is zero vector')
        else:
            return self / norm
        
    def length(self):
        """ Return the scalar length of a vector """
        return np.linalg.norm(self.vector_)

    @staticmethod
    def zero():
        """ Return the 2-dimensional zero vector """
        return Vector2(0.0, 0.0)
    
    @staticmethod
    def minimum(vec1, vec2):
        """ Return the minimum components of two vectors as a new Vector2 """
        return Vector2(np.minimum(tuple(vec1), tuple(vec2)))
    
    @staticmethod
    def maximum(vec1, vec2):
        """ Return the maximum components of two vectors as a new Vector2 """
        return Vector2(np.maximum(tuple(vec1), tuple(vec2)))


class Vector3(object):
    """ Utility functions for three dimensional vectors, which are represented
        as a list of three real values.
    """
    def __init__(self, *args):
        """ Construct a 3D vector, all zeroes by default. """
        self.vector_ = None
        if len(args) == 0:
            self.vector_ = np.zeros([3])
        elif len(args) == 1:
            if isinstance(args[0], (list, tuple)) and len(args[0]) == 3:
                self.vector_ = np.array([float(i) for i in args[0]])
            elif isinstance(args[0], np.ndarray):
                self.vector_ = args[0]
            elif isinstance(args[0], Vector3):
                self.vector_ = args[0].vector_
            elif isinstance(args[0], Vector2):
                self.vector_ = args[0].vector_
                self.vector_.append(0.0)
        elif len(args) == 3:
            self.vector_ = np.array([float(i) for i in args])

        if self.vector_ is None:
            raise ValueError("Invalid parameter %s" % str(args))
        elif self.vector_.size != 3 or self.vector_.ndim != 1:
            raise ValueError("Vector3 must be 3 float values")

    def __eq__(self, other):
        """ Check for equality of two vectors """
        return self.equal(other)

    def __add__(self, other):
        """ Add two vectors together and return the result as a new Vector3.
        """
        return Vector3(np.add(tuple(self), tuple(other)))
    
    def __sub__(self, other):
        """ Subtract one vector from another and return the result as a
            new Vector3.
        """
        return Vector3(np.subtract(tuple(self), tuple(other)))
    
    def __mul__(self, other):
        """ Multiply the components of a vector by either the components of
            another vector or a scalar and return the result as a new
            Vector3.
        """
        if isinstance(other, (float, int)):
            other = (other, other, other)

        return Vector3(np.multiply(self.vector_, tuple(other)))

    def __truediv__(self, other):
        """ Divide the components of a vector by either the components of
            another vector or a scalar and return the result as a new
            Vector3.
        """
        if isinstance(other, (float, int)) and other != 0.0:
            return Vector3(1.0 / other * self.vector_)
        else:
            return np.true_divide(self.vector_, tuple(other))
    
    def __div__(self, other):
        """ Divide the components of a vector by either the components of
            another vector or a scalar and return the result as a new
            Vector3.
        """
        return self.__truediv__(other)
    
    def __str__(self):
        """ Return a string representation of a Vector3 """
        return str(tuple(self))

    def __iter__(self):
        return (e for e in self.vector_)

    def __len__(self):
        return len(self.vector_)

    def index(self, i):
        """ Return a component of a Vector3 """
        return self.vector_[i]
    
    @property
    def x(self):
        """ The X component of a Vector3 """
        return self.vector_[0]

    @x.setter
    def x(self, v):
        """ Set the X component of a Vector3 """
        self.vector_[0] = v
    
    @property
    def y(self):
        """ The Y component of a Vector3 """
        return self.vector_[1]

    @y.setter
    def y(self, v):
        """ Set the Y component of a Vector3 """
        self.vector_[1] = v
    
    @property
    def z(self):
        """ The Z component of a Vector3 """
        return self.vector_[2]
    
    @z.setter
    def z(self, v):
        """ Set the Z component of a Vector3 """
        self.vector_[2] = v
    
    def equal(self, other, tolerance=GlyphUtility.tolerance):
        """ Check for equality between two vectors within an optional tolerance
        """
        return np.linalg.norm(np.subtract(tuple(self), tuple(other))) \
                <= tolerance
    
    def notEqual(self, other, tolerance=GlyphUtility.tolerance):
        """ Check for inequality between two vectors within an optional
            tolerance
        """
        return not self.equal(other, tolerance)
        
    def add(self, other):
        """ Add the components of two vectors and return the result as a new
            Vector3
        """
        return self + other
    
    def subtract(self, other):
        """ Subtract one vector from another and return the result as a new
            Vector3
        """
        return self - other
    
    def negate(self):
        """ Negate the component values and return the result as a new Vector3
        """
        return self * -1.0
    
    def scale(self, scalar):
        """ Scale the component values by a scalar and return the result as
            a new Vector3
        """
        return self * scalar
    
    def divide(self, scalar):
        """ Divide the component values by a scalar and return the result as
            a new Vector3
        """
        return self / scalar
    
    def multiply(self, scalar):
        """ Multiply the component values of a vector by the component values
            of another vector and return the result as a new Vector3
        """
        return self * scalar
    
    def cross(self, other):
        """ Return the cross product of two vectors as a new Vector3 """
        return Vector3(np.cross(self.vector_, tuple(other)))
    
    def dot(self, other):
        """ Return the scalar dot product of two vectors """
        return np.dot(self.vector_, tuple(other))
    
    def normalize(self):
        """ Return the normalized form of a vector as a new Vector3 """
        norm = np.linalg.norm(self.vector_)
        if norm == 0:
            raise ValueError('Vector is zero vector')
        else:
            return self / norm
    
    def length(self):
        """ Return the scalar length of a vector """
        return np.linalg.norm(self.vector_)
    
    def distanceToLine(self, pt, direction):
        """ Return the scalar distance from a vector to a line defined
            by a point and direction.
        """
        lineVec = Vector3(direction).normalize()
        ptVec = self - Vector3(pt)
        ptProj = lineVec * ptVec.dot(lineVec)
        return (ptVec - ptProj).length()
    
    @staticmethod
    def zero():
        """ Return the 3-dimensional zero vector """
        return Vector3(0.0, 0.0, 0.0)
    
    @staticmethod
    def minimum(vec1, vec2):
        """ Return the minimum components of two vectors as a new Vector3 """
        return Vector3(np.minimum(tuple(vec1), tuple(vec2)))
    
    @staticmethod
    def maximum(vec1, vec2):
        """ Return the maximum components of two vectors as a new Vector3 """
        return Vector3(np.maximum(tuple(vec1), tuple(vec2)))
    
    @staticmethod
    def affine(scalar, vec1, vec2):
        """ Return a new Vector3 that is the affine combination of two vectors
        """
        return Vector3(np.add(
            np.multiply(tuple(vec1), (scalar, scalar, scalar)),
            np.multiply(tuple(vec2), (1.0-scalar, 1.0-scalar, 1.0-scalar))))
    
    @staticmethod
    def barycentric(pt, vec1, vec2, vec3, clamp=False):
        """ Return a new Vector3 that has the barycentric coordinates of
            the given point in the frame of the given three vectors.
        """
        pt = Vector3(pt)
        v1 = Vector3(vec1)
        v2 = Vector3(vec2)
        v3 = Vector3(vec3)
        a = b = 0.0

        v12 = v2 - v1
        v13 = v3 - v1

        cross = v12.cross(v13)
        area = 0.5 * cross.length()

        if area == 0.0:
            mode = 23
            len12 = v12.length()
            len13 = v13.length()
            len23 = (v2 - v3).length()
            if len12 >= len13:
                if len12 >= len23:
                    mode = 11 if len12 == 0.0 else 12
            elif len13 >= len23:
                mode = 13

            if mode == 12:
                a = (pt - v2).length() / len12
                b = 1.0 - a
            elif mode == 13:
                a = (pt - v3).length() / len13
                b = 0.0
            elif mode == 23:
                a = 0.0
                b = (pt - v3).legnth() / len23
            else:  # mode = 11
                a = 1.0
                b = 0.0
        else:
            cross1 = (v2 - pt).cross(v3 - pt)
            a = 0.5 * cross1.length() / area
            if cross.dot(cross1) < 0.0:
                a = -a
            cross2 = (pt - v1).cross(v13)
            b = 0.5 * cross2.length() / area
            if cross.dot(cross2) < 0.0:
                b = -b

        a = 0.0 if clamp and a < 0.0 else a
        b = 0.0 if clamp and b < 0.0 else b
        b = 1.0 - a if clamp and (a + b) > 1.0 else b

        return Vector3((a, b, 1.0 - a - b))


class Quaternion(object):
    """ Utility functions for quaternions, which are represented as a list
        of four real values (x, y, z, and angle)
    """

    def __init__(self, axis=(0.0, 0.0, 0.0), angle=0, _quat=None):
        """ Construct a quaternion, with a degenerate axis and zero angle
            by default
        """
        if _quat is not None:
            self.quat_ = _quat
            # normalize
            self.quat_ /= np.sqrt(sum(k*k for k in self.quat_))
            self.angle_ = np.degrees(np.arccos(self.quat_[0]) * 2)
            self.axis_ = Vector3(self.quat_[1:4])
            self.axis_ = self.axis_ / np.sin(np.radians(self.angle_ / 2))
        elif len(axis) == 3:
            if not isinstance(axis, Vector3):
                axis = Vector3(axis)
            if axis.length() == 0.0 or angle == 0.0:
                self.quat_ = (1.0, 0.0, 0.0, 0.0)
                self.angle_ = 0.0
                self.axis_ = axis
            else:
                length = axis.length()

                # clamp the angle between -2*pi and +2*pi
                while angle < -360.0:
                    angle += 360.0
                while angle > 360.0:
                    angle -= 360.0

                if angle == 90.0 or angle == -270.0:
                    w = math.sqrt(0.5)
                    u = w / length
                elif angle == 180.0 or angle == -180.0:
                    w = 0.0
                    u = 1.0 / length
                elif angle == -90.0 or angle == 270.0:
                    w = math.sqrt(0.5)
                    u = -w / length
                else:
                    angle = angle / 2.0
                    w = np.cos(np.radians(angle))
                    u = np.sin(np.radians(angle)) / length

                # normalize, since that's what Glyph does
                self.quat_ = (w, u*axis.x, u*axis.y, u*axis.z)
                self.quat_ /= np.sqrt(sum(k*k for k in self.quat_))

                angleRadians = np.arccos(w) * 2.0
                self.angle_ = np.degrees(angleRadians)
                self.axis_ = Vector3(self.quat_[1:4]) / np.sin(angleRadians/2.0)
        else:
            raise ValueError("Invalid parameter")

        if len(self.quat_) != 4:
            raise ValueError("Quaternion must be list/tuple of 4 values")
        
    def __mul__(self, other):
        """ Return the quaternion product of two quaternions as a new Quaternion
        """
        if isinstance(other, (float, int)):
            return Quaternion(_quat=(self.quat_ * other))
        else:
            w0, x0, y0, z0 = self.quat_
            w1, x1, y1, z1 = other.quat_
            quatProduct = np.array(
                    [w1*w0 - x1*x0 - y1*y0 - z1*z0, \
                     w1*x0 + x1*w0 + y1*z0 - z1*y0, \
                     w1*y0 + y1*w0 + z1*x0 - x1*z0, \
                     w1*z0 + z1*w0 + x1*y0 - y1*x0])
            return Quaternion(_quat=quatProduct)
    
    def __truediv__(self, other):
        """ Return the quaternion divided by a scalar as a new Quaternion """
        other = (other, other, other, other)
        return Quaternion(_quat=(np.true_divide(self.quat_, other)))

    def __div__(self, other):
        return self.__truediv__(other)

    def __str__(self):
        """ Return the string form of a Quaternion as axis and angle """
        return "(%s, %f)" % (str(tuple(self.axis_)), self.angle_)

    def __iter__(self):
        return (e for e in (tuple(self.axis_) + (self.angle_,)))
    
    @property
    def axis(self):
        """ Return the rotation axis of a quaternion as a Vector3 """
        return self.axis_
    
    @property
    def angle(self):
        """ Return the scalar rotation angle in degrees """
        return self.angle_
    
    def equal(self, quat2):
        """ Compare two quaternions for equality """
        return np.array_equal(self.quat_, quat2.quat_)
    
    def notEquals(self, quat2):
        """ Compare two quaternions for inequality """
        return not self.equal(quat2)
    
    def rotate(self, quat2):
        """ Rotate a quaternion by another quaternion and return the result
            as a new Quaternion
        """
        return self * quat2
    
    def conjugate(self):
        """ Return the conjugate of a quaternion as a new Quaternion """
        return Quaternion(_quat=np.append(self.quat_[0], -1 * self.quat_[1:4]))
    
    def norm(self):
        """ Return the scalar normal for a quaternion """
        return np.linalg.norm(self.quat_)
    
    def inverse(self):
        """ Return the inverse of a quaternion as a new Quaternion """
        conj = self.conjugate().quat_
        norm2 = np.square(self.norm())
        return Quaternion(_quat=(conj / norm2))
        
    def normalize(self):
        """ Return the normalized version of a quaternion as a new Quaternion.
            Note: When constructed with an arbitrary axis and angle, a
            Quaternion is already normalized by default.
        """
        return self / self.norm()

    def asTransform(self):
        """ Compute and return a rotation Transform from a quaternion.
            This produces an exact Cartesian transformation when a quaternion
            axis is aligned with a Cartesian axis.
        """
        w, x, y, z = self.quat_
        q = x * x + y * y + z * z + w * w
        s = (2.0 / q) if q > 0.0 else 0.0
        xs = x * s
        ys = y * s
        zs = z * s
        wx = w * xs
        wy = w * ys
        wz = w * zs
        xx = x * xs
        xy = x * ys
        xz = x * zs
        yy = y * ys
        yz = y * zs
        zz = z * zs

        m = np.zeros([16])

        # In Glyph order

        # If along a Cartesian axis, snap to -1, 0, and 1 if within tolerance
        if (x == 0.0 and y == 0.0) or (x == 0.0 and z == 0.0) or \
                (y == 0.0 and z == 0.0):
            m[ 0] = _clamp(1.0 - (yy + zz), 1.0, -1.0, 1e-15)
            m[ 1] = _clamp(xy + wz, 1.0, -1.0, 1e-15)
            m[ 2] = _clamp(xz - wy, 1.0, -1.0, 1e-15)
            m[ 3] = 0.0
            m[ 4] = _clamp(xy - wz, 1.0, -1.0, 1e-15)
            m[ 5] = _clamp(1.0 - (xx + zz), 1.0, -1.0, 1e-15)
            m[ 6] = _clamp(yz + wx, 1.0, -1.0, 1e-15)
            m[ 7] = 0.0
            m[ 8] = _clamp(xz + wy, 1.0, -1.0, 1e-15)
            m[ 9] = _clamp(yz - wx, 1.0, -1.0, 1e-15)
            m[10] = _clamp(1.0 - (xx + yy), 1.0, -1.0, 1e-15)
            m[11] = 0.0
            m[12] = 0.0
            m[13] = 0.0
            m[14] = 0.0
            m[15] = 1.0
        else:
            m[ 0] = 1.0 - (yy + zz)
            m[ 1] = xy + wz
            m[ 2] = xz - wy
            m[ 3] = 0.0
            m[ 4] = xy - wz
            m[ 5] = 1.0 - (xx + zz)
            m[ 6] = yz + wx
            m[ 7] = 0.0
            m[ 8] = xz + wy
            m[ 9] = yz - wx
            m[10] = 1.0 - (xx + yy)
            m[11] = 0.0
            m[12] = 0.0
            m[13] = 0.0
            m[14] = 0.0
            m[15] = 1.0

        # Python order
        m = m.reshape((4,4)).transpose()

        return Transform(m)

class Plane(object):
    """ Utility functions for infinite planes, which are represented as a list
        of four plane coefficient values.
    """
    def __init__(self, **kargs):
        if len(kargs) == 1 and 'coeffs' in kargs:
            coeffs = kargs['coeffs']
            if not isinstance(coeffs, (list, tuple)) or len(coeffs) != 4:
                raise ValueError("Coefficients must be a list of 4 values")
            self.normal_ = Vector3(coeffs[0:3]).normalize()
            self.d_ = coeffs[3]
        elif len(kargs) == 2 and set(kargs.keys()) == set(('normal', 'origin')):
            self.normal_ = Vector3(kargs['normal']).normalize()
            self.d_ = self.normal_.dot(Vector3(kargs['origin']))
        elif len(kargs) == 3 and set(kargs.keys()) == set(('p1', 'p2', 'p3')):
            v0 = Vector3(kargs['p1'])
            v1 = Vector3(kargs['p2'])
            v2 = Vector3(kargs['p3'])
            self.normal_ = (((v1 - v0).normalize()).cross(
                      (v2 - v0).normalize())).normalize()
            self.d_ = self.normal_.dot(v0)
        elif len(kargs) == 4 and set(kargs.keys()) == set(('A', 'B', 'C', 'D')):
            self.normal_ = \
                    Vector3((kargs['A'], kargs['B'], kargs['C'])).normalize()
            self.d_ = kargs['D']
        else:
            raise ValueError('Plane must be initialized with coefficient ' +
                'list, point and normal, three points (p1, p2, p3), or ' +
                'coefficients (A, B, C, D)')

    def __iter__(self):
        return (e for e in (tuple(self.normal_) + (self.d_,)))

    def __str__(self):
        """ Return the string form of a plane represented as a tuple of
            the four plane coefficients
        """
        return str(tuple(self))

    def equation(self):
        """ Return the plane equation as tuple of the four plane coefficients
        """
        return tuple(self)

    @property
    def A(self):
        """ Return the A plane coefficient """
        return self.normal_.vector_[0]

    @property
    def B(self):
        """ Return the B plane coefficient """
        return self.normal_.vector_[1]

    @property
    def C(self):
        """ Return the C plane coefficient """
        return self.normal_.vector_[2]

    @property
    def D(self):
        """ Return the D plane coefficient """
        return self.d_

    def normal(self):
        """ Return the plane normal as a tuple """
        return tuple(self.normal_)

    def constant(self):
        """ Return the scalar plane constant """
        return self.d_

    def inHalfSpace(self, vec):
        """ Check if a point is in the positive half space of a plane """
        return self.normal_.dot(Vector3(vec)) >= self.d_

    def distance(self, vec):
        """ Return the positive scalar distance of a point to a plane """
        return math.fabs(self.normal_.dot(Vector3(vec)) - self.d_)

    def line(self, origin, direction):
        """ Return the intersection of a line represented as a point and
            distance to a plane. If the line does not intersect the plane,
            an exception is raised.
        """
        if not isinstance(origin, Vector3):
            origin = Vector3(origin)
        if not isinstance(direction, Vector3):
            direction = Vector3(direction).normalize()

        den = self.normal_.dot(direction)
        if den < 1e-10 and den > -1e-10:
            raise ValueError("Line does not intersect plane")

        s = (self.d_ - self.normal_.dot(origin)) / den

        return origin + direction * s

    def segment(self, p1, p2):
        """ Return the intersection of a line represented as two points
            to a plane as a new Vector3. If the segment does not
            intersect the plane, an exception is raised.
        """
        if not isinstance(p1, Vector3):
            p1 = Vector3(p1)
        if not isinstance(p2, Vector3):
            p2 = Vector3(p2)

        ndp1 = self.normal_.dot(p1)
        ndp2 = self.normal_.dot(p2)

        if ((ndp1 < self.d_ and ndp2 < self.d_) or \
                (ndp1 > self.d_ and ndp2 > self.d_)):
            raise ValueError("Segment does not intersect plane")

        return self.line(p1, tuple(p2 - p1))

    def project(self, pt):
        """ Return the closest point projection of a point onto a plane
            as a new Vector3
        """
        if not isinstance(pt, Vector3):
            pt = Vector3(pt)
        return pt + self.normal_ * (self.d_ - pt.dot(self.normal_))


class Extents(object):
    """ Utility functions for extent boxes, which are represented as a
        list of two vectors (the min and max of the box).
    """ 
    def __init__(self, *args):
        """ Construct an extent box with the given min/max or None.
            Extents((xmin, ymin, zmin), (xmax, ymax, zmax))
            Extents(Vector3, Vector3)
            Extents(xmin, ymin, zmin, xmax, ymax, zmax)
        """
        self.box_ = None
        if len(args) == 1:
            if isinstance(args[0], Extents):
                self.box_ = args[0].box_
            else:
                self.box_ = np.array(args[0])
        elif len(args) == 2:
            if isinstance(args[0], Vector3) and isinstance(args[1], Vector3):
                self.box_ = np.array([args[0].vector_, args[1].vector_])
            elif isinstance(args[0], (list, tuple, np.ndarray)) and \
                    isinstance(args[1], (list, tuple, np.ndarray)):
                self.box_ = np.array([args[0], args[1]])
            else:
                raise ValueError("Invalid argument %s" % str(args))
        elif len(args) == 3:
            self.box_ = np.array([args, args])
        elif len(args) == 6:
            self.box_ = np.array(args)
            self.box_.shape = [2,3]
        elif len(args) != 0:
            raise ValueError("Invalid argument %s" % str(args))

        if self.box_ is not None:
            if self.box_.size == 3 and self.box_.ndim == 1:
                self.box_ = np.array([self.box_, self.box_])
            if self.box_.size != 6 or self.box_.ndim != 2:
                raise ValueError("Extent box must be 2x3 matrix")
            elif not np.less_equal(self.box_[0], self.box_[1]).all():
                raise ValueError("Min must be less than or equal to Max")

    def __repr__(self):
        return str(self.box_)

    def __iter__(self):
        if self.box_ is not None:
            return (e for e in self.box_.flatten())

    def __eq__(self, other):
        """ Check for equality of two extent boxes """
        if self.box_ is not None:
            try:
                return np.array_equal(self.box_, other.box_)
            except:
                return np.array_equal(self.box_, other)
        else:
            return False

    def minimum(self):
        """ Return the minimum corner point of an extent box """
        if self.box_ is None:
            raise ValueError("Self is empty")
        return self.box_[0]

    def maximum(self):
        """ Return the maximum corner point of an extent box """
        if self.box_ is None:
            raise ValueError("Self is empty")
        return self.box_[1]

    def isEmpty(self):
        """ Check if an extent box is empty """
        return self.box_ is None

    def diagonal(self):
        """ Return the length of the diagonal of an extent box """
        if self.box_ is not None:
            return np.linalg.norm(self.box_[0] - self.box_[1])
        else:
            return 0.0

    def enclose(self, pt):
        """ Return a new Extents that encloses the given point """
        if isinstance(pt, Vector3):
            pt = pt.vector_
        if self.box_ is None:
            return Extents([np.array(pt), np.array(pt)])
        else:
            return Extents([np.minimum(self.box_[0], pt),
                np.maximum(self.box_[1], pt)])

    def expand(self, value):
        """ Return a new Extents box that is expanded by the given amount
            at both minimum and maximum corners
        """
        return Extents([self.box_[0] - value, self.box_[1] + value])

    def isIntersecting(self, other):
        """ Return true if two extent boxes intersect or share a corner, edge
            or face
        """
        if isinstance(other, (list, tuple)):
            other = Extents(other)
        elif not isinstance(other, Extents):
            raise ValueError("Invalid argument")
        elif self.box_ is None:
            return False

        return (max(self.box_[0, 0], other.box_[0, 0]) <= \
                min(self.box_[1, 0], other.box_[1, 0])) and \
                (max(self.box_[0, 1], other.box_[0, 1]) <= \
                min(self.box_[1, 1], other.box_[1, 1])) and \
                (max(self.box_[0, 2], other.box_[0, 2]) <= \
                min(self.box_[1, 2], other.box_[1, 2]))

    def isInside(self, pt, tol=0.0):
        """ Return true if a point is within an extent box, within an
            optional tolerance
        """
        if self.box_ is None:
            return False

        pt = tuple(pt)

        return (self.box_[0, 0] + tol) <= pt[0] and \
                (self.box_[1, 0] - tol) >= pt[0] and \
                (self.box_[0, 1] + tol) <= pt[1] and \
                (self.box_[1, 1] - tol) >= pt[1] and \
                (self.box_[0, 2] + tol) <= pt[2] and \
                (self.box_[1, 2] - tol) >= pt[2]

    def translate(self, offset):
        """ Return a new Extents object that is translated by the given offset
        """
        if self.box_ is None:
            raise ValueError("Self is empty")
        elif isinstance(offset, Vector3):
            offset = tuple(offset)

        return Extents([np.add(self.box_[0], tuple(offset)),
            np.add(self.box_[1], tuple(offset))])

    def rotate(self, quat):
        """ Return a new Extents object that is rotated by the given Quaternion
        """
        if self.box_ is None:
            raise ValueError("Self is empty")
        elif not isinstance(quat, Quaternion):
            raise ValueError("quat is not a Quaternion")

        xform = quat.asTransform()
        result = Extents()

        obox = self.box_

        for i in (0,1):
            for j in (0,1):
                for k in (0,1):
                    p = xform.apply((obox[i, 0], obox[j, 1], obox[k, 2]))
                    result = result.enclose(p)

        return result

    def center(self):
        """ Return the center point of the extent box
        """
        if self.box_ is None:
            raise ValueError("Self is empty")
        return (self.box_[0] / 2.0) + (self.box_[1] / 2.0)

    def minimumSide(self):
        """ Return the length of the shortest side of the box
        """
        if self.box_ is None:
            raise ValueError("Self is empty")
        return min(self.box_[0, 0]-self.box_[1, 0],
                self.box_[0, 1]-self.box_[1, 1],
                self.box_[0, 2], self.box_[1, 2])

    def maximumSide(self):
        """ Return the length of the longest side of the box
        """
        if self.box_ is None:
            raise ValueError("Self is empty")
        return max(self.box_[0, 0]-self.box_[1, 0],
                self.box_[0, 1]-self.box_[1, 1],
                self.box_[0, 2], self.box_[1, 2])


class Transform(object):
    """ Utility functions for transform matrices, which are represented
        as a list of sixteen real values. The matrix is represented
        in a column-first order, which matches the order used in Glyph.
    """
    @staticmethod
    def identity():
        """ Return an identity Transform """
        return Transform(list(np.eye(4)))

    def __init__(self, matrix=None):
        """ Construct a Transform from an optional set of 16 real values.
            If an argument is not supplied, set to the identity matrix.
        """
        if matrix is None:
            self.xform_ = Transform.identity().xform_
            return
        elif isinstance(matrix, Quaternion):
            self.xform_ = matrix.asTransform().xform_
            return
        elif isinstance(matrix, Transform):
            self.xform_ = matrix.xform_
            return
        elif isinstance(matrix, np.matrix):
            self.xform_ = matrix.A
            return
        elif isinstance(matrix, np.ndarray):
            self.xform_ = matrix
            if matrix.size == 16:
                self.xform_ = self.xform_.reshape((4,4))
            return
        elif not isinstance(matrix, (list, tuple)):
            raise ValueError("Invalid argument")

        # Assume that the incoming transformation matrix is in Glyph order
        # (column-first), and transpose to row-first for mathematical
        # operations
        if len(matrix) == 4:
            # Assume a list/tuple of 4 lists/tuples
            self.xform_ = np.array(matrix).transpose()
        elif len(matrix) == 16:
            self.xform_ = np.array(matrix).reshape((4,4)).transpose()
        else:
            raise ValueError("Invalid xform matrix")

    def __eq__(self, other):
        """ Compare equality of two transforms """
        if isinstance(other, Transform):
            return np.array_equal(self.xform_, other.xform_)
        elif isinstance(other, (list, tuple, np.matrix, np.ndarray)):
            return np.array_equal(self.xform_, Transform(other).xform_)
        else:
            raise ValueError("Invalid argument")

    def __str__(self):
        """ Return the string representation of a Transform, as a one-
            dimensional array of floats in column-wise order
        """
        return str(tuple(self))

    def __iter__(self):
        # account for transposed glyph notation
        return (e for e in self.xform_.transpose().flatten())

    @property
    def matrix(self):
        return list(self)

    def element(self, i, j):
        """ Get an element of a transform matrix with i, j in Glyph order """
        return self.xform_.transpose()[i, j]

    @staticmethod
    def translation(offset):
        """ Return a new Transform that is a translation by the given offset """
        return Transform.identity().translate(offset)

    def translate(self, offset):
        """ Return a new Transform that adds translation to an existing
            Transform
        """
        if not isinstance(offset, Vector3):
            offset = Vector3(offset)
        col = np.dot(self.xform_, np.array(tuple(offset.vector_) + (1.0,)))
        xf = np.array(self.xform_)
        np.put(xf, [3, 7, 11, 15], col)
        return Transform(xf)

    @staticmethod
    def rotation(axis, angle, anchor=None):
        """ Return a new Transform that is a rotation by the given angle
            about the given axis at the (optional) given anchor point
        """
        return Transform.identity().rotate(axis, angle, anchor)

    def rotate(self, axis, angle, anchor=None):
        """ Return a new Transform that adds rotation to a Transform by
            the given angle about the given axis at the (optional) given
            anchor point
        """
        if not isinstance(axis, Vector3):
            axis = Vector3(axis)
        if anchor is not None:
            if not isinstance(anchor, Vector3):
                anchor = Vector3(anchor)
            axform = Transform.translation(anchor)
            axform = axform.rotate(axis, angle)
            axform = axform.translate(Vector3()-anchor)
            return Transform(np.dot(self.xform_, axform.xform_))

        # handle Cartesian rotations
        cartesian = False
        if axis.x == 0.0 and axis.y == 0.0:
            # Axis is 0 0 Z
            cartesian = True
            ct1 = 0
            ct2 = 5
            pst = 1
            nst = 4
            if axis.z > 0.0:
                angle = -angle
        elif axis.x == 0.0 and axis.z == 0.0:
            # Axis is 0 Y 0
            cartesian = True
            ct1 = 0
            ct2 = 10
            pst = 2
            nst = 8
            if axis.y < 0.0:
                angle = -angle
        elif axis.y == 0.0 and axis.z == 0.0:
            # Axis X 0 0
            cartesian = True
            ct1 = 5
            ct2 = 10
            pst = 9
            nst = 6
            if axis.x < 0.0:
                angle = -angle

        if cartesian:
            absAngle = math.fmod(math.fabs(angle), 360.0)

            if absAngle == 90.0:
                ca = 0.0
                sa = 1.0
            elif absAngle == 180.0:
                ca = -1.0
                sa = 0.0
            elif absAngle == 270.0:
                ca = 0.0
                sa = -1.0
            else:
                ca = math.cos(math.radians(absAngle))
                sa = math.sin(math.radians(absAngle))

            if angle < 0.0:
                sa = -sa

            mat = np.eye(4).flatten()
            mat[ct1] = ca
            mat[nst] = -sa
            mat[pst] = sa
            mat[ct2] = ca

            rxform = Transform(mat)
        else:
            rxform = Quaternion(axis, angle).asTransform()

        return Transform(np.dot(self.xform_, rxform.xform_))

    @staticmethod
    def scaling(scale, anchor=None):
        """ Return a new Transform that is a scale by the given factor
            (which can be a scalar or a three-dimensional vector) about
            an optional anchor point
        """
        return Transform.identity().scale(scale, anchor)

    def scale(self, scale, anchor=None):
        """ Return a new Transform that adds scaling to a Transform by
            the given factor (which can be a scalar or a three-
            dimensional vector) about an optional anchor point
        """
        if isinstance(scale, (float, int)):
            scale = ((float(scale), float(scale), float(scale)))
        else:
            scale = tuple(scale)

        if anchor is not None:
            if not isinstance(anchor, Vector3):
                anchor = Vector3(anchor)
            axform = Transform.translation(anchor)
            axform = axform.scale(scale)
            axform = axform.translate(Vector3()-anchor)
            return Transform(np.dot(self.xform_, axform.xform_))

        # transpose to make slicing easier
        xf = self.xform_.transpose()
        xf[0][0:3] = np.multiply(xf[0][0:3], scale)
        xf[1][0:3] = np.multiply(xf[1][0:3], scale)
        xf[2][0:3] = np.multiply(xf[2][0:3], scale)
        xf = xf.transpose()

        return Transform(xf)

    @staticmethod
    def calculatedScaling(anchor, start, end, tol=0.0):
        """ Return a transform matrix that scales a given point from
            one location to another anchored at a third point
        """
        if isinstance(anchor, (list, tuple)):
            anchor = Vector3(anchor)
        if isinstance(start, (list, tuple)):
            start = Vector3(start)
        if isinstance(end, (list, tuple)):
            end = Vector3(end)

        fac = Vector3()

        da0 = start - anchor
        da1 = end - anchor
        fac.x = 1.0 if math.fabs(da0.x) < tol else (da1.x / da0.x)
        fac.y = 1.0 if math.fabs(da0.y) < tol else (da1.y / da0.y)
        fac.z = 1.0 if math.fabs(da0.z) < tol else (da1.z / da0.z)

        end1 = ((start - anchor) * fac) + anchor

        da1 = end1 - anchor
        fac.x = 1.0 if math.fabs(da0.x) < tol else (da1.x / da0.x)
        fac.y = 1.0 if math.fabs(da0.y) < tol else (da1.y / da0.y)
        fac.z = 1.0 if math.fabs(da0.z) < tol else (da1.z / da0.z)

        return Transform.scaling(fac, anchor)

    @staticmethod
    def ortho(left, right, bottom, top, near, far):
        """ Return an orthonormal view Transform from a view frustum """
        if (left == right):
            raise ValueError("left and right plane constants cannot be equal")
        if (bottom == top):
            raise ValueError("bottom and top plane constants cannot be equal")
        if (near == far):
            raise ValueError("near and far plane constants cannot be equal")

        irml = 1.0 / (right - left)
        itmb = 1.0 / (top - bottom)
        ifmn = 1.0 / (far - near)

        mat = np.eye(4).flatten()

        # these are in Python order
        mat[0] = 2.0 * irml
        mat[5] = 2.0 * itmb
        mat[10] = -2.0 * ifmn
        mat[3] = -(right + left) * irml
        mat[7] = -(top + bottom) * itmb
        mat[11] = -(far + near) * ifmn

        return Transform(mat)

    @staticmethod
    def perspective(left, right, bottom, top, near, far):
        """ Return a perspective view Transform from a view frustum """
        if (left == right):
            raise ValueError("left and right plane constants cannot be equal")
        if (bottom == top):
            raise ValueError("bottom and top plane constants cannot be equal")
        if (near == far):
            raise ValueError("near and far plane constants cannot be equal")

        irml = 1.0 / (right - left)
        itmb = 1.0 / (top - bottom)
        ifmn = 1.0 / (far - near)

        mat = np.zeros([16])

        # these are in Glyph order
        mat[0] = 2.0 * near * irml
        mat[5] = 2.0 * near * itmb
        mat[8] = (right + left) * irml
        mat[9] = (top + bottom) * itmb
        mat[10] = -(far + near) * ifmn
        mat[11] = -1.0
        mat[14] = -2.0 * far * near * ifmn

        # Python order
        mat = mat.reshape((4,4)).transpose()

        return Transform(mat)

    @staticmethod
    def mirroring(normal, distance):
        """ Return a new Transform that mirrors about a plane given
            by a normal vector and a scalar distance
        """
        return Transform.identity().mirror(normal, distance)

    @staticmethod
    def mirrorPlane(plane):
        """ Return a new Transform that mirrors about a given plane """
        return Transform.identity().mirror(plane.normal_, plane.d_)

    def mirror(self, normal, distance):
        """ Return a new Transform that adds mirroring about a plane
            given by a normal vector and a scalar distance
        """
        if not isinstance(normal, Vector3):
            normal = Vector3(normal)
        normal = normal.normalize()

        # These are in Glyph order
        mat = np.zeros([16])
        mat[ 0] = 1.0 - 2.0 * normal.x * normal.x
        mat[ 1] = -2.0 * normal.y * normal.x
        mat[ 2] = -2.0 * normal.z * normal.x
        mat[ 3] = 0.0
        mat[ 4] = -2.0 * normal.x * normal.y
        mat[ 5] = 1.0 - 2.0 * normal.y * normal.y
        mat[ 6] = -2.0 * normal.z * normal.y
        mat[ 7] = 0.0
        mat[ 8] = -2.0 * normal.x * normal.z
        mat[ 9] = -2.0 * normal.y * normal.z
        mat[10] = 1.0 - 2.0 * normal.z * normal.z
        mat[11] = 0.0
        mat[12] = 2.0 * normal.x * distance
        mat[13] = 2.0 * normal.y * distance
        mat[14] = 2.0 * normal.z * distance
        mat[15] = 1.0

        # transpose back to Python order
        mat = mat.reshape((4,4)).transpose()

        return Transform(np.dot(self.xform_, mat))

    @staticmethod
    def stretching(anchor, start, end):
        """ Return a new Transform that is a stretching transform.
            If the vector defined by the start and end points is
            orthogonal to the vector defined by the start and anchor
            points, the transform is undefined and the matrix will
            be set to the identity matrix.
        """
        return Transform.identity().stretch(anchor, start, end)

    def stretch(self, anchor, start, end):
        """ Return a new Transform that adds stretching to a Transform.
            If the vector defined by the start and end points is
            orthogonal to the vector defined by the start and anchor
            points, the transform is undefined and the matrix will
            be set to the identity matrix.
        """
        if not isinstance(anchor, Vector3):
            anchor = Vector3(anchor)
        if not isinstance(start, Vector3):
            start = Vector3(start)
        if not isinstance(end, Vector3):
            end = Vector3(end)

        aToStart = start - anchor
        aToEnd = end - anchor

        sDir = (end - start).normalize()

        if math.fabs(aToStart.normalize().dot(sDir)) >= 0.01:
            factor = (aToEnd.dot(sDir) / aToStart.dot(sDir)) - 1.0
            # These are in Glyph order
            mat = np.eye(4).flatten()
            mat[ 0] = factor * sDir.x * sDir.x
            mat[ 1] = factor * sDir.y * sDir.x
            mat[ 2] = factor * sDir.z * sDir.x
            mat[ 4] = factor * sDir.x * sDir.y
            mat[ 5] = factor * sDir.y * sDir.y
            mat[ 6] = factor * sDir.z * sDir.y
            mat[ 8] = factor * sDir.x * sDir.z
            mat[ 9] = factor * sDir.y * sDir.z
            mat[10] = factor * sDir.z * sDir.z

            # transpose back to Python order to apply translation
            mat = mat.reshape((4,4)).transpose()

            axform = Transform.translation(Vector3()-anchor)
            mat = np.dot(mat, axform.xform_)
            axform = Transform.translation(anchor)
            mat = np.dot(axform.xform_, mat)

            # Glyph order
            mat = mat.transpose().flatten()
            # mat = np.ravel(mat)
            mat[ 0] += 1.0
            mat[ 5] += 1.0
            mat[10] += 1.0
            mat[12] -= anchor.x
            mat[13] -= anchor.y
            mat[14] -= anchor.z

            # Python order
            mat = mat.reshape((4,4)).transpose()

            return Transform(np.dot(self.xform_, mat))
        else:
            return Transform(self.xform_)
                
    def apply(self, vec):
        """ Return a new Vector3 that is transformed from a given point """
        # apply transform to a point
        if not isinstance(vec, Vector3):
            vec = Vector3(vec)
        rw = np.array(tuple(vec.vector_) + (1.0,))
        rw = np.dot(self.xform_, rw)
        vec = np.array(rw[0:3])
        if rw[3] != 0.0:
            vec = vec / rw[3]
        return Vector3(vec)

    def applyToDirection(self, direct):
        """ Return a a new Vector3 that is a transformed direction vector.

            This differs from apply as follows:
            
            When transforming a point by a 4x4 matrix, the point is
            represented by a vector with X, Y, and Z as the first 3
            components and 1 as the fourth component.  This allows
            the point to pick up any translation component in the matrix.
            This method represents the direction as a vector with 0 as
            the fourth component.  Since a direction can be thought of
            as the difference between two points, a zero fourth component
            is the difference between two points that have 1 as the fourth
            component.
        """
        # apply transform to a direction vector, as opposed to a point
        if not isinstance(direct, Vector3):
            direct = Vector3(direct)
        rw = np.array(tuple(direct.vector_) + (0.0,))
        rw = np.dot(self.xform_, rw)
        direct = np.array(rw[0:3])
        if rw[3] != 0.0:
            direct = direct / rw[3]
        return Vector3(direct)

    def applyToNormal(self, normal):
        """ Return a new Vector3 that is a transformed normal vector.
            A normal vector is transformed by multiplying the normal
            by the transposed inverse matrix.
        """
        if not isinstance(normal, Vector3):
            normal = Vector3(normal)
        rw = np.array(tuple(normal.vector_) + (0.0,))
        rw = np.dot(np.linalg.inv(self.xform_).transpose(), rw)
        normal = np.array(rw[0:3])
        return Vector3(normal).normalize()

    def applyToPlane(self, plane):
        """ Return a new Plane that is transformed plane """
        o = self.apply(plane.project((0, 0, 0)))
        n = self.applyToNormal(plane.normal_)
        return Plane(origin=o, normal=n)

    def __mul__(self, other):
        """ Return a the multiplication of self and either another Transform
            (a new Transform) or a Vector3 (a new, transformed Vector3)
        """
        if isinstance(other, Vector3):
            return self.apply(other)
        elif isinstance(other, Transform):
            return Transform(np.dot(self.xform_, other.xform_))
        else:
            raise ValueError("Invalid argument")

    def multiply(self, other):
        """ Return a new Transform that is the multiplication of self
            and another Transform. This is equivalent to 'self * other'.
        """
        if isinstance(other, Transform):
            return self * other
        else:
            raise ValueError("Invalid argument")

    def determinant(self):
        """ Return the scalar determinant of the transformation matrix """
        return np.linalg.det(self.xform_)

    def transpose(self):
        return Transform(self.xform_.transpose())

    def inverse(self):
        return Transform(np.linalg.inv(self.xform_))

def _clamp(v, high, low, tol=0.0):
    """ Clamp a value to a range, or zero if within tolerance """
    if v > high-tol:
        return high
    elif v < low+tol:
        return low
    elif v > -tol and v < tol:
        return 0.0
    else:
        return v

#############################################################################
#
# This file is licensed under the Cadence Public License Version 1.0 (the
# "License"), a copy of which is found in the included file named "LICENSE",
# and is distributed "AS IS." TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE
# LAW, CADENCE DISCLAIMS ALL WARRANTIES AND IN NO EVENT SHALL BE LIABLE TO
# ANY PARTY FOR ANY DAMAGES ARISING OUT OF OR RELATING TO USE OF THIS FILE.
# Please see the License for the full text of applicable terms.
#
#############################################################################



#! /usr/bin/env python

#############################################################################
#
# (C) 2021 Cadence Design Systems, Inc. All rights reserved worldwide.
#
# This sample script is not supported by Cadence Design Systems, Inc.
# It is provided freely for demonstration purposes only.
# SEE THE WARRANTY DISCLAIMER AT THE BOTTOM OF THIS FILE.
#
#############################################################################

"""
This module provides a Python-like interface to Glyph.

GlyphAPI provides the service of automatically converting Python commands
that look like static method calls into Glyph static actions.
"""
from pointwise import GlyphClient
from pointwise import GlyphError

from .glyphobj import GlyphObj, GlyphVar
from .utilities import *

class GlyphAPI(object):
    """ This class provides access to Glyph static actions.  """

    def __init__(self, glyph_client):
        """ Initialize a GlyphAPI object from a connected GlyphClient.
        """
        self.glf = glyph_client
 
        # Acquire the list of all valid Glyph class names to seed the
        # name list
        if self.glf is not None and self.glf.is_connected():
            self._names = self.glf.eval(
                    'pw::Application getAllCommandNames')
        else:
            self._names = []
            raise GlyphError('', 'Not connected')

    def __getattr__(self, name):
        """ Create a GlyphObj object as needed for a Glyph class to invoke
            its static actions.
        """
        glfFunction = 'pw::' + name
        glfObj = None

        if glfFunction in self._names:
            glfObj = GlyphObj(glfFunction, self.glf)
            setattr(self, name, glfObj)

        return glfObj

#############################################################################
#
# This file is licensed under the Cadence Public License Version 1.0 (the
# "License"), a copy of which is found in the included file named "LICENSE",
# and is distributed "AS IS." TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE
# LAW, CADENCE DISCLAIMS ALL WARRANTIES AND IN NO EVENT SHALL BE LIABLE TO
# ANY PARTY FOR ANY DAMAGES ARISING OUT OF OR RELATING TO USE OF THIS FILE.
# Please see the License for the full text of applicable terms.
#
#############################################################################



#! /usr/bin/env python

#############################################################################
#
# (C) 2021 Cadence Design Systems, Inc. All rights reserved worldwide.
#
# This sample script is not supported by Cadence Design Systems, Inc.
# It is provided freely for demonstration purposes only.
# SEE THE WARRANTY DISCLAIMER AT THE BOTTOM OF THIS FILE.
#
#############################################################################

from .glyph_client import GlyphClient, GlyphError

#############################################################################
#
# This file is licensed under the Cadence Public License Version 1.0 (the
# "License"), a copy of which is found in the included file named "LICENSE",
# and is distributed "AS IS." TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE
# LAW, CADENCE DISCLAIMS ALL WARRANTIES AND IN NO EVENT SHALL BE LIABLE TO
# ANY PARTY FOR ANY DAMAGES ARISING OUT OF OR RELATING TO USE OF THIS FILE.
# Please see the License for the full text of applicable terms.
#
#############################################################################



#! /usr/bin/env python

from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='pointwise-glyph-client',
      version='2.0.11',
      description='Glyph client in Python with Python-like API to Pointwise Glyph Server',
      url='http://github.com/pointwise/GlyphClientPython',
      install_requires=['numpy'],
      packages=['pointwise', 'pointwise.glyphapi'],
      classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 2.7'],
      author='Cadence Design Systems, Inc.',
      author_email='pw-support@cadence.com')



#! /usr/bin/env python

#############################################################################
#
# (C) 2021 Cadence Design Systems, Inc. All rights reserved worldwide.
#
# This sample script is not supported by Cadence Design Systems, Inc.
# It is provided freely for demonstration purposes only.
# SEE THE WARRANTY DISCLAIMER AT THE BOTTOM OF THIS FILE.
#
#############################################################################

"""
This module provides a Python-like interface to Glyph.

GlyphObj provides the service of automatically converting Python
function calls into the equivalent Glyph command string and creates
Python objects to represent Glyph classes and objects.

GlyphVar provides a convenient interface for representing Tcl variables
as Python objects for Glyph commands that accept Tcl variable names as
arguments. These are typically used to return some optional Glyph result
value.
"""

from pointwise import GlyphClient, GlyphError
import keyword, json, re

class GlyphVar(object):
    """ A GlyphVar object wraps a Tcl variable name that may be passed to
        a Glyph action, which is typically used to return some value
        that is not the return value from the action. A variable name
        can be supplied; if not, a temporary Tcl variable name will be
        generated.

        Tcl:
            pw::Display selectEntities -selectionmask $mask ents

        where 'ents' is a Tcl variable name

        Python:
            m = pw.Display.createSelectionMask(requireConnector="Dimensioned")
            entsVar = GlyphVar()
            pw.Display.selectEntities(entsVar, selectionMask=m)
            cons = entsVar["Connectors"]

        where 'entsVar' is a GlyphVar object containing the 'ents' result value
    """

    def __init__(self, varname=None, value=None):
        """ Initializes GlyphVar object with its initial value """
        self.value = value
        self.varname_ = varname

    def __getitem__(self, item):
        """ Allows the GlyphVar to be subscriptable if applicable """
        return self.value[item]

    def __setitem__(self, key, value):
        """ Allows the GlyphVar to have values assigned to it if applicable """
        self.value[key] = value

    def __iter__(self):
        """ Allows the GlyphVar to be iterable """
        return (e for e in self.value)


class GlyphObj(object):
    """ This class implements a wrapper around Glyph Objects. In this context,
        a Glyph object can be a Glyph function (instance) for instance actions,
        or a Glyph class for static actions.
    """

    # pre-compiled regular expression matching Glyph function (object) names
    _functionRE = re.compile(r"::pw::[a-zA-Z]+_\d+( |$)")

    def __init__(self, function, glf):
        """ Glyph object constructor from a Glyph function name. Glyph wrappers
            are typically created automatically from the result of other Glyph
            actions.

            Args:
                function - the glyph function name or JSON dict with
                    command element

                glf - the GlyphClient connection in use

            Example:
                con = GlyphObj("pw::Connector_1", pw)
        """
        self.glf = glf
        if isinstance(function, dict):
            self._function = function['command']
            self._type = function['type']
        else:
            self._function = function
            if GlyphObj._functionRE.match(function):
                self._type = glf.eval("%s getType" % function)
            else:
                self._type = None

    def __enter__(self):
        """ Glyph objects derived from pw::Mode or pw::Examine can be
            used as context managers. For modes, the context will
            automatically issue the 'mode end' action upon exit, unless
            an exception occurs in which case 'mode abort' will be issued.
            Examine, the Glyph examine object will be destroyed when the
            context exits.

            Ex.
                with pw.Application.begin("Create") as creator:
                    con = pw.Connector()
                    ...

                with pw.Examine("BlockJacobian") as exam:
                    exam.addEntity(blk)
                    exam.examine()
                    ...
        """
        self._isMode = self._isExamine = self._open = False

        try:
            x = self.glf.eval(self._function + " isOfType pw::Examine")
            self._isExamine = bool(int(x))
        except Exception:
            self._isExamine = False

        if not self._isExamine:
            try:
                x = self.glf.eval(self._function + " isOfType pw::Mode")
                self._isMode = bool(int(x))
            except Exception:
                self._isMode = False

        if self._isMode:
            self._open = True

            def _end():
                try:
                    self.glf.eval(self._function + " end")
                    self._open = False
                except:
                    pass

            setattr(self, "end", _end)
        elif self._isExamine:
            self._open = True

            def _delete():
                try:
                    self.glf.eval(self._function + " delete")
                    self._open = False
                except:
                    pass

            setattr(self, "delete", _delete)
        else:
            raise GlyphError("", "A GlyphObj can be used as a context " + \
                    "manager for mode and examine objects only.")

        return self


    def __exit__(self, ex_type, ex_val, ex_tb):
        """ Exit context, ending or aborting mode or deleting examine object
            as needed
        """
        try:
            if self._open:
                if self._isMode:
                    if ex_type is None:
                        self.glf.eval(self._function + " end")
                    else:
                        self.glf.eval(self._function + " abort")
                elif self._isExamine:
                    self.glf.eval(self._function + " delete")
                self._open = False
                return False
        finally:
            del self._isMode
            del self._isExamine
            del self._open


    # Create Glyph object when called
    def __call__(self, *args):
        """ Creates GlyphObj from the result of the glyph command
                '<glyph-class> create'. This is valid only for
                objects that represent Glyph classes that implement
                the 'create' action.

            Args:
                args - a list of positional arguments to be passed to
                    the Glyph 'create' action

            Returns:
                GlyphObj object that was created

            Example:
                con = pw.Connector()
                exa = pw.Examine("ConnectorLengthI")
        """
        gvars = {}
        kwargs = {}
        cmd = self._buildGlyphCmd(self._function, 'create', args, kwargs, gvars)
        result = self._runGlyphCmd(cmd)
        return result


    def __str__(self):
        """ Returns a string representation of the Glyph object. """
        name = self.glf.eval("if [%s isOfType pw::Entity] { %s getName }" % \
                (self._function, self._function))
        return "%s (%s)" % (self._function, name)


    # Method for comparing GlyphObj objects            
    def __eq__(self, other):
        """ Compare two Glyph objects for equality. Glyph objects are
            equal if their Glyph functions refer to the same Glyph object.

            Returns:
                True if GlyphObj objects refer to the same Glyph object.
        """
        cmd = [self._function, 'equals', other._function]
        return bool(self._runGlyphCmd(cmd))


    # Add hash support so GlyphObj objects can be used in dictionaries
    def __hash__(self):
        return hash(self._function)


    @property
    def glyphType(self):
        """ Return the Glyph type of a GlyphObj. """
        return self._type


    @staticmethod
    def _toGlyph(value, varDict):
        """ Helper function to process a Glyph result value. """
        result = None
        if isinstance(value, GlyphVar):
            # Create variable in Glyph and store its reference name in varDict
            if value.varname_ is None:
                value.varname_ = '_TMPvar_%d' % len(varDict)
            result = value.varname_
            varDict[value] = result
        elif isinstance(value, GlyphObj):
            # Substitute the Glyph function name
            result = value._function
        elif isinstance(value, str):
            # Handle string arguments
            result = value
        else:
            try:
                # See if the object is iterable
                it = iter(value)
                result = [GlyphObj._toGlyph(v, varDict) for v in it]
            except:
                result = value
                pass

        return result


    @staticmethod
    def _buildGlyphCmd(function, action, args, kwargs, gvars):
        """ Helper function that builds command list to be converted to JSON. 

            Args: 
                function: the Glyph object or class on which the action
                    is to be run

                action: the Glyph action to be run on the Glyph object or class

                args - list of non-keyword (positional) arguments for the
                    Glyph action. Positional arguments always appear at
                    the end of a Glyph action, after all command switches.

                kwargs - list of keyword (switch) arguments for the Glyph
                    action. Switch arguments always appear immediately
                    after the action name. The following rules apply
                    when keyward arguments are present:

                    - the keyword is added to the Glyph command as a switch
                        (e.g., '-orient')
                    - if the keyword argument value is None, the switch
                        has no arguments in the action
                    - if the keyword argument value is a scalar (non-list)
                        value, the value is added right after the switch
                        (e.g., '-orient Best')
                    - if the keyword argument value is a list, a Tcl
                        list is added right after the switch, UNLESS the
                        keyword itself has a trailing underscore. In that
                        case, the list arguments are added one at a time
                        after the switch. For example,

                        Python               Tcl
                        arg=[s1, s2]   => -arg [list s1 s2]
                        arg_=[s1, s2]  => -arg s1 s2

                gvars: dictionary used to keep track of Glyph variables. Key
                    is GlyphVar object and value is Glyph variable name

            Returns: 
                JSON serializable list to be sent to _runGlyphCmd function
        """
        cmd = []

        # Some glyph commands are python keywords, and so must be passed with
        # a trailing underscore
        if keyword.iskeyword(action[:-1]) and action[-1] == '_':
            action = action[:-1]

        cmd.append(function)
        cmd.append(action)

        # Add keyword args as flags first (Glyph standard command notation)
        for flag, value in kwargs.items():
            # special convention for flags that accept multiple arguments:
            #    foo  = [a, b]           => -foo {a b}
            #    foo_ = [a, b]           => -foo a b
            #    foo  = [[a, b], [c, d]] => -foo {{a b} {c d}}
            #    foo_ = [[a, b], [c, d]] => -foo {a b} {c d}

            # special convention for keywords with boolean values:
            #   foo=True    => -foo
            #   foo_=True   => -foo true
            #   foo=False   => (flag excluded from command)
            #   foo_=False  => -foo false
            
            flatten = flag.endswith('_')

            if flatten:
                flag = flag[:-1]

            if not flatten and isinstance(value, bool) and bool(value) == False:
                continue

            gval = GlyphObj._toGlyph(value, gvars)

            cmd.append("-%s" % flag)

            if flatten:
                cmd += gval
            elif not isinstance(value, bool):
                cmd.append(gval)

        # Add positional args (never flattened) to the command
        cmd += [GlyphObj._toGlyph(arg, gvars) for arg in args]

        return cmd


    @staticmethod
    def _isGlyphFunction(arg):
        """ Helper function to determine if an object or list of objects can
            be converted to a GlyphObj

            Args:
                arg - element to check if it is something returned from JSON

            Returns:
                True if element can be converted to a GlyphObj
        """

        if isinstance(arg, dict):
            # JSON result from Glyph action
            return all(k in arg for k in ('command', 'type'))
        elif isinstance(arg, str) and GlyphObj._functionRE.match(arg):
            # Glyph function name
            return True
        else:
            return False


    def _runGlyphCmd(self, cmd):
        """ Helper function to run a Glyph action that is represented as
            a list of string command arguments

            Args:
                cmd - A JSON serializable list that represents the Glyph action

            Returns:
                Return value from Glyph action, converted to a GlyphObj if
                    possible
        """

        # convert the command list to a JSON string, execute the command, and 
        # convert the result back from a JSON string to a Python list
        result = json.loads(self.glf.command(json.dumps(cmd)))

        # If the result is a list, convert any JSON dict elements to GlyphObj,
        # allowing for nested lists
        if isinstance(result, list):
            result = [self._toPythonObj(k) for k in result]
        else:
            result = self._toPythonObj(result)

        return result


    def __getattr__(self, action):
        """ Create a method on demand that mimics some Glyph action.
            Note: Glyph action names that conflict with Python reserved words
            must be appended with an underscore. E.g.
            
            pw.Grid.import_(filename)
        """
        def _action_(*args, **kwargs):
            """ Used to generate Glyph calls 'on the fly'. A JSON command is
                created based on the GlyphObj action called. The value that
                is returned from the server is then converted back to GlyphObj
                object if applicable.

                Returns:
                    Whatever Glyph command would return as GlyphObj object
                    type if applicable
            """
            function = self._function  

            # dict of GlyphObj (Tcl variable)
            gvars = {}

            # create command token list
            cmd = self._buildGlyphCmd(function, action, args, kwargs, gvars)
            try:
                result = self._runGlyphCmd(cmd)
                # evaluate Glyph variable contents
                if gvars:
                    for pyVar, tclVarName in gvars.items():
                        pyVar.value = self._evalTclVar(tclVarName)
                    self._delTclVar(gvars)
                if self._isMode and action in ['end', 'abort']:
                    self._open = False
                elif self._isExamine and action == 'delete':
                    self._open = False
                return result
            except GlyphError:
                raise

        setattr(self, action, _action_)
        return _action_ 


    def _delTclVar(self, gvars):
        """ Helper function to delete (unset) a dictionary of Glyph (Tcl)
            variables

            Args:
                gvars - dictionary of Glyph variables with key as
                    GlyphVar object and value as Glyph name

            Returns:
                None
        """
        _unset = "if [info exists %s] { unset %s }"
        cmd = [(_unset % (var, var)) for var in gvars.values()]
        self.glf.eval('; '.join(cmd))


    def _toPythonObj(self, tclArg):
        """ Helper function to convert a Tcl string to a Python string or
            GlyphObj object

            Args:
                tclArg - item to convert

            Returns:
                Python string or GlyphObj
        """
        result = tclArg
        # handle nested lists
        if isinstance(result, list):
            result = [self._toPythonObj(k) for k in result]
        elif GlyphObj._isGlyphFunction(tclArg):
            result = GlyphObj(tclArg, self.glf)
        elif isinstance(tclArg, str):
            tclArg = tclArg.strip()
            if len(tclArg) > 0:
                try:
                    result = int(tclArg)
                except ValueError:
                    try:
                        result = float(tclArg)
                    except ValueError:
                        result = tclArg
            else:
                result = None
        return result


    def _toPythonList(self, tclArg):
        """ Helper function to convert a Tcl list to a Python list, converting
            Glyph function names to GlyphObj objects where possible. The
            input list can be as deeply nested as needed, and the resulting
            Python list will be nested to the same depth. Tcl strings that can
            be converted to numeric Python objects will be so converted.

            Args:
                tclArg - Tcl list or value to be converted to Python list of
                    string or GlyphObj values. A Tcl list must be in the
                    form "{ v1 v2 v3 }". Nested Tcl lists are handled.

                Tcl                            Python
                { { 0 0 1 } { 0 1 0 } }     => [[0, 0, 1], [0, 1, 0]]

                { 0.0 1.0 ::pw::Surface_1 } => [0.0, 1.0, gobj]
                    where gobj = GlyphObj("::pw::Surface_1")

                {::pw::Surface_1 ::pw::Connector_1 } => [gobj1, gobj2]
                    where gobj1 = GlyphObj("::pw::Surface_1")
                          gobj2 = GlyphObj("::pw::Connector_1")

            Returns:
                Python list where every element is a string, number, or
                GlyphObj object
        """
        out = []
        cache = [out]
        element = ''
        escape = False
        for char in tclArg:
            if escape:
                if char not in ["\\", "{", "}", "[", "]", "$"]:
                    raise ValueError("Incorrect escape character %s" % char)
                element += char
                escape = False
            elif char == "\\":
                escape = True
            elif char in [" ", "\t", "\r", "\n"]:
                element = self._toPythonObj(element)
                if element is not None:
                    cache[-1].append(element)
                element = ''
            elif char == "{":
                level = []
                cache[-1].append(level)
                cache.append(level)
            elif char == "}":
                if len(cache) < 2:
                    raise ValueError("Close bracket without opening bracket.")
                element = self._toPythonObj(element)
                if element is not None:
                    cache[-1].append(element)
                cache.pop()
                element = ''
            else:
                element += char

        element = self._toPythonObj(element)
        if element is not None:
            cache[-1].append(element)

        if len(cache) != 1:
            raise ValueError("Mismatched brackets.")

        if len(out) == 1:
            out = out[0]
        return out


    def _evalTclVar(self, tclVarName):
        """ Evaluate a Tcl variable and convert to Python. A Tcl array will
            be converted to a Python dictionary. All other Tcl values will
            be converted to a list of Python values.
            
            Note: scalar Tcl values will be returned as a list with a
            single element. Values that are Glyph functions (objects)
            will be converted to GlyphObj objects.

            Args:
                tclVarName - Glyph variable name to be evaluated and
                    stored in a Python dictionary or list

            Returns: 
                A Python Dictionary
        """
        # ask Tcl interpreter if the variable exists
        isvar = int(self.glf.eval("info exists %s" % tclVarName))
        if not isvar:
            return None

        # ask Tcl interpreter if the variable is an array
        isdict = int(self.glf.eval("array exists %s" % tclVarName))
        if isdict:
            # Convert Tcl array to Python dict
            tcl = self.glf.eval("array get %s" % tclVarName)
        else:
            # Convert Tcl list (or single value) to a Python list
            tcl = self.glf.eval("lrange $%s 0 end" % tclVarName)

        result = self._toPythonList(tcl)

        if isdict:
            # make each dictionary value a list as needed
            i = iter(result)
            result = dict(zip(i, i))
            for key, value in result.items():
                if not isinstance(value, list):
                    result[key] = [value]

        return result

#############################################################################
#
# This file is licensed under the Cadence Public License Version 1.0 (the
# "License"), a copy of which is found in the included file named "LICENSE",
# and is distributed "AS IS." TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE
# LAW, CADENCE DISCLAIMS ALL WARRANTIES AND IN NO EVENT SHALL BE LIABLE TO
# ANY PARTY FOR ANY DAMAGES ARISING OUT OF OR RELATING TO USE OF THIS FILE.
# Please see the License for the full text of applicable terms.
#
#############################################################################
